#lang racket
;;; adt-turtle-hard-music.rkt pour Racket 
;;; (jpr, PF2)
(provide turtle% TURTLES repeat)

;;; Version de la tortue comme objet avec l'API objet de Racket !

(require graphics/graphics)     ; la librairie graphics de la collection graphics
(open-graphics)                 ; chargement de la librairie
(define SIZE 250)
(define SIZE/2 (quotient SIZE 2))
(define twin (open-viewport "adt-turtle-hard-sound" SIZE SIZE))
(define tr-segment (draw-line twin))    ; doc de draw-line pour ajouter la couleur !!!

(define (%turtlepoint->posn x y)        ; changement de repere
  (make-posn (+ SIZE/2 x) (- SIZE/2 y)))   ; repere au centre, unite 1 pixel

(define (real-modulo a b)    ; a mod b, a et b reels, b > 0, resultat dans [0,b[
  (define (descendre a)      ; methode un peu bovine mais qui devrait suffire...
    (if (< a b) a (descendre (- a b))))
  (define (monter a)
    (if (> a 0) a (monter (+ a b))))
  (cond ((>= a b) (descendre a))
        ((>= a 0) a)
        (else (monter a))))  ; beurk beurk et beurk !

(define PI/180 (/ pi 180))
(define 180/PI (/ 180 pi))

(define (degrees->radians angle-en-degres)
  (* angle-en-degres PI/180)) 

(define (radians->degrees angle-en-radians)
  (* angle-en-radians 180/PI))

;;; une p'tite boucle sympa qui manque en Scheme (macro ajoutable) et en
;;; Python (hum, non ajoutable)... A placer dans utils.rkt

(define-syntax repeat            ; la boucle (repeat n e1 e2 ...), avec n entier >= 0
  (syntax-rules ()
    ((repeat n e1 e2 ...) (for ([i (in-range n)]) e1 e2 ...))))

;;; sonorisation
(require rsound)
(define ps (make-pstream))
(define BOING (rs-read "boing.wav"))
(define HELPME (rs-read "help-me.wav"))

;;; On commence a s'occuper des tortues ici

(define TURTLES '())     ; la liste globale de toutes les tortues

(define turtle%
  (class object%
    (init-field (XCOR 0) (YCOR 0) (CAP 0) (PENDOWN? #t) (SLEEP 0.06))
    (define/public (init pos cap . clear?)    ; (send t init pos cap) ou (send t init pos cap #t)
      (when (or (null? clear?) (equal? #t (car clear?)))
        ((clear-viewport twin)))
      (pen-up)
      (set-position pos)
      (set-heading cap)
      (pen-down))
    (define/public (pen-up)
      (set! PENDOWN? #f))
    (define/public (pen-down)
      (set! PENDOWN? #t))
    (define/public (set-heading angle-en-degres)
      (set! CAP (real-modulo angle-en-degres 360)))
    (define/public (heading)
      CAP)
    (define/public (set-position L)          ; L == (x y) est un turtlepoint
      (let ((x (car L)) (y (cadr L)))
        (when PENDOWN? 
          (tr-segment (%turtlepoint->posn XCOR YCOR) (%turtlepoint->posn x y)))
        (set! XCOR x)
        (set! YCOR y)
        (sleep SLEEP)))
    (define/public (position)
      (list XCOR YCOR))
    (define/public (xcor)
      XCOR)
    (define/public (ycor)
      YCOR)
    (define/public (right angle-en-degres)
      (set-heading (+ CAP angle-en-degres)))
    (define/public (left angle-en-degres)
      (set-heading (- CAP angle-en-degres)))
    (define/public (forward dist)
      (define rad (degrees->radians CAP))
      (set-position (list (+ XCOR (* dist (sin rad)))
                          (+ YCOR (* dist (cos rad))))))
    (define/public (back dist)
      (forward (- dist)))
    (define/public (within-bounds?)
      (and (< (- SIZE/2) XCOR SIZE/2) (< (- SIZE/2) YCOR SIZE/2)))
    (define/public (come-back-into-bounds!)
      (pstream-play ps BOING)
      (cond ((< XCOR (- SIZE/2)) (set-position `(,(- SIZE/2) ,YCOR)))
            ((> XCOR SIZE/2) (set-position `(,SIZE/2 ,YCOR)))
            ((< YCOR (- SIZE/2)) (set-position `(,XCOR ,(- SIZE/2))))
            (else (set-position `(,XCOR ,SIZE/2))))
      (toward '(0 0)))    ; hum
    (define/public (toward L)    ; force la tortue a regarder vers le point L = (x y)
      (let ((x (car L)) (y (cadr L)) (pos (position)))
        (let ((xp (car pos)) (yp (cadr pos)))
          (let ((dx (- xp x)) (dy (- yp y)) (angle '?))
            (if (= dx 0) 
                (set! angle (if (> dy 0) -90 90))
                (set! angle (radians->degrees (atan (/ dy dx)))))
            (set! angle (- angle))
            (when (> dx 0) (set! angle (+ angle 180)))  ; atan defini entre 0 et 180
            (set-heading (+ angle 90))))))
    (set! TURTLES (cons this TURTLES))
    (super-new)))

(define musical-turtle%
  (class turtle%
    (inherit-field XCOR YCOR)
    (define (xcor->freq x)    ; [-SIZE/2,SIZE/2] --> [300,700], de 300Hz à gauche jusqu'à 700Hz à droite
      (+ 450 (* 500 x (/ SIZE))))
    (define (ycor->lowfreq y) ; [-SIZE/2,SIZE/2] --> [0.5,8], de 0.5Hz en bas jusqu'à 8Hz en haut
      (+ 4.25 (* y (/ 7.5 SIZE))))
    (define signal
      (network ()
               (freq = (xcor->freq XCOR))
               (a <= sawtooth-wave freq)
               (lfo <= sine-wave (ycor->lowfreq YCOR))
               (out = (* 0.3 a lfo))))
    (super-new)  ; le bon moment d'initialiser XCOR et YCOR !
    (signal-play signal)))

(define lea (new musical-turtle%))

(with-handlers ([exn:break? (lambda (e) (stop))])    ; bouton "Stopper"
  (for ([i (in-range 2000)])
       (send lea forward 4)
       (if (not (send lea within-bounds?))
           (send lea come-back-into-bounds!)
           (send lea left (- (random 60) 30)))))
(stop)


