#lang racket
;;; adt-turtle-hard.rkt pour Racket 
;;; La version objet du fichier adt-turtle.rkt
;;; (solution PF2, OCT 2015)
(provide turtle% TURTLES repeat distance)

;;; -------------- PRELIMINAIRES ----------------------------------------------------

(require graphics/graphics)     ; la librairie graphics simplifiee, de la collection graphics
(open-graphics)                 ; chargement de la librairie
(define SIZE 500)
(define SIZE/2 (quotient SIZE 2))
(define twin (open-viewport "Turtlegraphics" SIZE SIZE))
(define tr-segment (draw-line twin))       ; doc de draw-line pour ajouter la couleur !!!

(define (%turtlepoint->posn x y)           ; changement de repere
  (make-posn (+ SIZE/2 x) (- SIZE/2 y)))   ; repere au centre, unite 1 pixel

(define (real-modulo a b)    ; a mod b, a et b reels, b > 0, resultat dans [0,b[
  (define (descendre a)      ; methode un peu bovine mais qui devrait suffire...
    (if (< a b) a (descendre (- a b))))
  (define (monter a)
    (if (> a 0) a (monter (+ a b))))
  (cond ((>= a b) (descendre a))
        ((>= a 0) a)
        (else (monter a))))  ; beurk beurk et beurk !

(define PI/180 (/ pi 180))
(define 180/PI (/ 180 pi))

(define (degrees->radians angle-en-degres)
  (* angle-en-degres PI/180)) 

(define (radians->degrees angle-en-radians)
  (* angle-en-radians 180/PI))

;;; une p'tite boucle sympa qui manque en Scheme et en
;;; Python (gasp, impossible)... A placer dans utils.rkt

(define-syntax repeat            ; la boucle (repeat n e1 e2 ...), avec n entier >= 0
  (syntax-rules ()
    ((repeat n e1 e2 ...) (for ([i (in-range n)]) e1 e2 ...))))

;;; ----------------------------------------------------------------
;;; Version de la tortue vraiment objet avec l'API objet de Racket !
;;; Bonus : une methode d'instance distance qui permet de demander
;;; a une tortue sa distance a une autre tortue. Mais j'exporte
;;; aussi une fonction (distance t1 t2)...

(define TURTLES '())     ; la liste globale de toutes les tortues, si l'on veut parler au peuple !

(define turtle%
  (class object%
    (init-field (XCOR 0) (YCOR 0) (CAP 0) (PENDOWN? #t))
    (define/public (init pos cap [clear? #t])    ; (send t init pos cap) ou (send t init pos cap #f)
      (when clear? ((clear-viewport twin)))
      (pen-up)
      (set-position pos)    ; <==> (send this set-position pos)
      (set-heading cap)     ; En effet dans le texte d'une classe - et comme en Java - une methode est aussi une fonction !
      (pen-down))           ; mais a l'exterieur de la classe, il faut envoyer un message avec send !
    (define/public (pen-up)
      (set! PENDOWN? #f))
    (define/public (pen-down)
      (set! PENDOWN? #t))
    (define/public (set-heading angle-en-degres)
      (set! CAP (real-modulo angle-en-degres 360)))
    (define/public (heading)
      CAP)
    (define/public (set-position L)          ; L == (x y) est un turtlepoint
      (let ((x (car L)) (y (cadr L)))
        (when PENDOWN? 
          (tr-segment (%turtlepoint->posn XCOR YCOR) (%turtlepoint->posn x y)))
        (set! XCOR x)
        (set! YCOR y)))
    (define/public (position)
      (list XCOR YCOR))
    (define/public (xcor)
      XCOR)
    (define/public (ycor)
      YCOR)
    (define/public (right angle-en-degres)
      (set-heading (+ CAP angle-en-degres)))
    (define/public (left angle-en-degres)
      (set-heading (- CAP angle-en-degres)))
    (define/public (forward dist)
      (set-position (list (+ XCOR (* dist (sin (degrees->radians CAP))))
                          (+ YCOR (* dist (cos (degrees->radians CAP)))))))
    (define/public (back dist)
      (forward (- dist)))
    (define/public (toward L)    ; force la tortue a regarder vers le point L = (x y)
      (let ((x (car L)) (y (cadr L)) (pos (position)))
        (let ((xp (car pos)) (yp (cadr pos)))
          (let ((dx (- xp x)) (dy (- yp y)) (angle '?))
            (if (= dx 0) 
                (set! angle (if (> dy 0) -90 90))
                (set! angle (radians->degrees (atan (/ dy dx)))))
            (set! angle (- angle))
            (when (> dx 0) (set! angle (+ angle 180)))  ; atan est defini entre 0 et 180
            (set-heading (+ angle 90))))))
    (define/public (distance other)             ; je ne prends pas la distance euclidienne !
      (+ (abs (- XCOR YCOR))
         (abs (- (send other ycor) (send other ycor)))))
    (set! TURTLES (cons this TURTLES))
    (super-new)))

(define (distance t1 t2)                        ; je ne prends pas la distance euclidienne usuelle !
  (+ (abs (- (send t1 xcor) (send t2 xcor)))    ; celle-ci est plus rapide lorsque la distance est petite
     (abs (- (send t1 ycor) (send t2 ycor)))))  ; (les cercles deviennent des losanges)...




