#lang racket
;;; tp5.rkt

(require "utils.rkt")              ; avec show, while, etc. Adaptez...

"************************* COURS 5 ***************************"

(define (new-pile)                 ; le constructeur de pile vide
  (define L '())
  (define (this methode . Largs)   ; construction de l'objet courant (self en Python, this en Racket et Java)
    (case methode
      ((vide?) (null? L))
      ((sommet) (if (null? L) (error "Pile vide !") (car L)))
      ((empiler) (set! L (cons (car Largs) L)))
      ((depiler) (if (null? L) (error "Pile vide !") (set! L (cdr L))))
      (else (error "Méthode inconnue" methode))))
  this)     ; le resultat est l'objet this, fonction analysant le message

(define P (new-pile))
(show P)
(show (P 'empiler 3))
(show (P 'empiler 5))
(show (P 'empiler 7))
(show (P 'vide?))
(show (P 'sommet))
(show (P 'depiler))
(show (P 'depiler))
(show (P 'sommet))
(printf "--------------\n")

(define (new-point2d x0 y0)    ; le constructeur du point (x0,y0)
  (let ((x x0) (y y0))
    (define (this method . Largs)
      (case method
        ((getX) x)
        ((getY) y)
        ((setX) (set! x (car Largs)))
        ((setY) (set! y (car Largs)))
        ((reset) (set! x x0) (this 'setY y0))     ; <---
        ((toString) (format "point2d[x=~a,y=~a]" x y))
        (else (error "Methode inconnue" method))))
    this))

(define a (new-point2d 1 2))
(show a)
(show (a 'setY (+ (a 'getY) 1)))        ; en Java:  a.setY(a.getY()+1);
(show (a 'toString))                    ; en Java:  a.toString()
(show (a 'reset))
(show (a 'toString))

"LES OBJETS HARD"

(define pile%
  (class object%                      ; une sous-classe de object%
    (public vide? sommet empiler depiler)
    (define L '())
    (define (vide?)
      (null? L))
    (define (sommet)
      (if (null? L) (error "Pile vide !") (car L)))
    (define (empiler x)
      (set! L (cons x L)))
    (define (depiler)
      (if (null? L) (error "Pile vide !") (set! L (cdr L))))
    (super-new)))

(define P1 (new pile%))
(show P1)
(show (send P1 vide?))
(show (send P1 empiler 3))
(show (send P1 empiler 5))
(show (send P1 sommet))
(show (send P1 depiler))
(show (send P1 sommet))
(printf "--------------\n")

(define forme<%>
  (interface () pos move))

(define nbPoints 0)    ; variable static en Java...

(define point2d%
  (class* object% (forme<%>)   ; implemente l'interface forme<%>, donc s'engage a programmer pos et move !
    (init-field (x 0) (y 0))
    (define/public (pos) (list x y))
    (define/public (move dx dy)
      (set! x (+ x dx)) (set! y (+ y dy)))
    (define/public (toString)
      (concat-infos))
    (define (concat-infos)          ; private !
      (format "point2d[~a,~a]" x y))
    (set! nbPoints (add1 nbPoints))
    (super-new)))

(define a1 (new point2d% (x 10) (y 20)))
(define b1 (new point2d% (y 20)))
(show a1)
(show b1)
(show (object? a1))
(show (is-a? a1 point2d%))
(show (send a1 pos))
(show (send a1 move 2 5))
(show (send a1 toString))
(printf "--------------\n")

(define colored-point2d%
  (class point2d%
    (init-field (color "black"))
    (inherit-field x y)
    (define/public (getColor)
      color)
    (define/override (toString)    
      (string-append "colored-" (super toString) "-" color))
    (define/public (state)
      (list x y color))    ; color ⇔ (send this getColor)
    (super-new)))

(define P2 (new colored-point2d% (x 5) (color "red")))
(show P2)
(show (send P2 pos))
(show (send P2 getColor))
(show (send P2 toString))
(show (and (is-a? P2 object%) 
           (is-a? P2 point2d%) 
           (is-a? P2 colored-point2d%) 
           (is-a? P2 forme<%>)))

"************************* TP 5 ***************************"

(define (protected-value x passwd)       ; x est la valeur originale
  (define value x)                       ; value est la valeur courante
  (define (this method . Largs)
    (case method
      ((getValue) value)
      ((setValue) (printf "Gimme a password : ")
                  (if (equal? (read) passwd)
                      (set! value (car Largs))
                      (printf "Access Denied !\n")))
      ((reset) (this 'setValue x))
      (else (error 'protected-value "Unknown method : ~a" method))))
  this)

(define V (protected-value 2015 'unsa))
(printf "Le password est : unsa\n")
(show (V 'getValue))
(show (V 'setValue 2021))     ; en Objective-C sur Mac/iPhone on ecrirait :    [V setValue: 2021]
(show (V 'getValue))

"TORTUE HARD"

(require "adt-turtle-hard.rkt")    ; pour turtle%, repeat et distance

(define (cercles n)                ; n cercles en rotation
  (define jojo (new turtle%))
  (define (cercle)                 ; polygone regulier...
    (repeat 36 (send* jojo (forward 10) (left 10))))      ; une cascade de messages...
  (define a (/ 360.0 n))
  (repeat n (cercle) (send jojo left a)))

(cercles 10)
(sleep 6)

; 4 tortues a chaque coin du canvas. Chaque tortue poursuit la suivante !
; J'utilise un vecteur VT de 4 tortues (cours 11, ou livre PCPS page 265). Avantage
; sur une liste : acces O(1) a chaque element. Peut-être dans les projets ?...

(define (poursuite)     
  (define VT (build-vector 4 (lambda (i) (new turtle%))))   ; un vecteur de 4 tortues
  (send (vector-ref VT 0) init '(-250 250) 90 #t)           ; #f optionnel : sans effacer la fenetre !
  (send (vector-ref VT 1) init '(250 250) 180 #f)
  (send (vector-ref VT 2) init '(250 -250) -90 #f)
  (send (vector-ref VT 3) init '(-250 -250) 0 #f)
  (while (> (distance (vector-ref VT 0) (vector-ref VT 1)) 4)
    (for ([i (in-range 4)])
      (define ti (vector-ref VT (modulo i 4)))
      (define ti+1 (vector-ref VT (modulo (+ i 1) 4)))
      (send ti toward (send ti+1 position))
      (send ti forward 1)))
  (printf "C'EST FINI !\n"))

(poursuite)

; Ah ! Il parait que chaque tortue parcourt
; une distance egale au cote du canvas (?)...


