#lang racket
;;; Racket 6.2.1
;;; Fichier d'utilitaires pour les TP d'option PF2

(provide while show 
         empiler pile-vide pile-vide? depiler sommet 
         arbre racine feuille? fg fd operateur?)

(define-syntax while
  (syntax-rules ()
    ((while test e1 e2 ...)
     (let ()
       (define (iter)
         (if (not test)
             (void)
             (begin e1 e2 ... (iter))))
       (iter)))))

;Exemple d'utilisation :
#|
 (let ((x 10) (s 0))
   (while (> x 0)
     (set! s (+ s x))
     (set! x (- x 1)))
   (printf "1 + 2 + 3 + ... + 10 = ~a\n" s))
|#

(define-syntax show
  (syntax-rules ()
    ((show expr) (begin (printf "? ~s\n" 'expr) 
                        (define res expr)
                        (when (not (void? res))
                          (printf "--> ~s\n" res))))))

; Exemples d'utilisation :
; (show (sqrt 5))
; (show (print 5))

;;; ------------------ utils pour le livre PCPS -----------------------------

; le type abstrait "arbre binaire d'expression algebrique" 
; Toutes les operations sont O(1)

(define (arbre r Ag Ad)    
  (list r Ag Ad))

(define (racine A)
  (if (feuille? A)
      (error "racine : pas de racine pour une feuille : " A)
      (car A)))

(define (fg A)
  (if (feuille? A)
      (error "pas de fg pour une feuille : " A)
      (cadr A)))

(define (fd A)
  (if (feuille? A)
      (error "pas de fd pour une feuille : " A)
      (caddr A)))

(define (feuille? obj)
  (or (number? obj) 
      (and (symbol? obj) (not (operateur? obj)))))

(define (operateur? obj)
  (member obj '(+ * - /)))

; le type abstrait "pile fonctionnelle" pour le livre PCPS.
; Toutes les operations sont O(1)

(define (pile-vide) 
  empty)

(define (pile-vide? pile) 
  (empty? pile))

(define (empiler x pile)
  (cons x pile)) 

(define (sommet pile)
  (if (empty? pile)
      (error "sommet : Pile vide !")
      (car pile)))

(define (depiler pile)
  (if (empty? pile)
      (error "depiler : Pile vide !")
      (cdr pile)))




